#!/usr/bin/env python3
# (c) 2024, Cantab Research Ltd.
"""
Command-line interface
"""

import json
import logging
import ssl
import sys
from dataclasses import dataclass
from socket import gaierror

import httpx
from websockets.exceptions import WebSocketException

from speechmatics_flow.cli_parser import parse_args
from speechmatics_flow.client import WebsocketClient
from speechmatics_flow.exceptions import ConversationError
from speechmatics_flow.models import (
    AudioSettings,
    ConversationConfig,
    ServerMessageType,
    Interaction,
    ConnectionSettings,
)
from speechmatics_flow.templates import TemplateOptions

LOGGER = logging.getLogger(__name__)


@dataclass
class Transcripts:
    user_transcript: str = ""
    agent_transcript: str = ""


def get_log_level(verbosity):
    """
    Returns the appropriate log level given a verbosity level.

    :param verbosity: Verbosity level.
    :type verbosity: int

    :return: The logging level (eg. logging.INFO).
    :rtype: int

    :raises SystemExit: If the given verbosity level is invalid.
    """
    try:
        log_level = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbosity]

        return log_level
    except KeyError as error:
        key = int(str(error))
        raise SystemExit(
            f"Only supports 2 log levels eg. -vv, you are asking for " f"-{'v' * key}"
        ) from error


def get_connection_settings(args):
    """
    Helper function which returns a ConnectionSettings object based on the
    command line options given to the program.

    :param args: Keyword arguments, typically from the command line.
    :type args: dict

    :return: Settings for the WebSocket connection.
    :rtype: models.ConnectionSettings
    """
    auth_token = args.get("auth_token")
    url = args.get("url")
    settings = ConnectionSettings(
        url=url,
        auth_token=auth_token,
        generate_temp_token=args.get("generate_temp_token") == "true",
    )

    if args.get("buffer_size") is not None:
        settings.message_buffer_size = args["buffer_size"]

    if args.get("ssl_mode") == "insecure":
        settings.ssl_context.check_hostname = False
        settings.ssl_context.verify_mode = ssl.CERT_NONE
    elif args.get("ssl_mode") == "none":
        settings.ssl_context = None

    return settings


def get_conversation_config(
    args,
):
    """
    Helper function which returns a ConversationConfig object based on the
    command line options given to the program.

    :param args: Keyword arguments probably from the command line.
    :type args: Dict

    :return: Settings for the Flow engine.
    :rtype: models.ConversationConfig
    """

    config = {}
    # First, get configuration from the config file if provided.
    if args.get("config_file"):
        with open(args["config_file"], encoding="utf-8") as config_file:
            config = json.load(config_file)

    if config.get("conversation_config"):
        config = config["conversation_config"]

    # Command line arguments override values from config file
    if assistant := args.get("assistant"):
        config["template_id"] = TemplateOptions.get(assistant)

    return ConversationConfig(**config)


def get_audio_settings(args):
    """
    Helper function which returns an AudioSettings object based on the command
    line options given to the program.

    Args:
        args (dict): Keyword arguments, typically from the command line.

    Returns:
        models.AudioSettings: Settings for the audio stream
            in the connection.
    """
    settings = AudioSettings(
        sample_rate=args.get("sample_rate"),
        chunk_size=args.get("chunk_size"),
        encoding=args.get("raw"),
    )
    return settings


# pylint: disable=too-many-arguments,too-many-statements
def add_printing_handlers(
    api,
    transcripts,
    print_json=False,
):
    """
    Adds a set of handlers to the websocket client which print out transcripts
    as they are received. This includes partials if they are enabled.

    Args:
        api (client.WebsocketClient): Client instance.
        transcripts (Transcripts): Allows the transcripts to be concatenated to
            produce a final result.
        print_json (bool, optional): Whether to print json transcript messages.
    """
    escape_seq = "\33[2K" if sys.stdout.isatty() else ""

    def convert_to_txt(message):
        if print_json:
            print(json.dumps(message))
            return
        transcript = message["metadata"]["transcript"]
        return transcript.replace("<sb> ", "").replace("</sb>", "").replace("<sb>", "")

    def transcript_handler(message):
        plaintext = convert_to_txt(message)
        if plaintext:
            sys.stdout.write(f"{escape_seq}{plaintext}\n")
        transcripts.user_transcript += plaintext

    def partial_transcript_handler(message):
        plaintext = convert_to_txt(message)
        if plaintext:
            sys.stderr.write(f"{escape_seq}{plaintext}\r")

    def prompt_handler(message):
        if print_json:
            print(json.dumps(message))
            return
        new_response = message["content"]
        new_plaintext_response = new_response.replace("<sb> ", "").replace("</sb> ", "")
        if new_plaintext_response:
            sys.stdout.write(f"{escape_seq}{new_plaintext_response}\n")
        transcripts.user_transcript += new_plaintext_response

    api.add_event_handler(ServerMessageType.ResponseStarted, prompt_handler)
    api.add_event_handler(ServerMessageType.AddTranscript, transcript_handler)
    api.add_event_handler(
        ServerMessageType.AddPartialTranscript, partial_transcript_handler
    )


# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
def main(args=None):
    """
    Main entrypoint.

    :param args: command-line arguments; defaults to None in which
            case arguments will be retrieved from `sys.argv` (this is useful
            mainly for unit tests).
    :type args: List[str]
    """
    if not args:
        args = vars(parse_args())

    logging.basicConfig(level=get_log_level(args["verbose"]))
    LOGGER.info("Args: %s", args)

    try:
        flow_main(args)
    except (KeyboardInterrupt, ValueError, ConversationError, KeyError) as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(f"{type(error).__name__}: {error}")
    except FileNotFoundError as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(
            f"FileNotFoundError: {error.strerror}: '{error.filename}'."
            + " Check to make sure the filename is spelled correctly, and that the file exists."
        )
    except httpx.HTTPStatusError as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(error.response.text)
    except httpx.HTTPError as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(f"httpx.HTTPError: An unexpected http error occurred. {error}")
    except ConnectionResetError as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(
            f"ConnectionResetError: {error}.\n\nThe most likely reason for this is that the client "
            + "has been configured to use SSL but the server does not support SSL. "
            + "If this is the case then try using --ssl-mode=none"
        )
    except (WebSocketException, gaierror) as error:
        LOGGER.info(error, exc_info=True)
        sys.exit(
            f"WebSocketError: An unexpected error occurred in the websocket: {error}.\n\n"
            + "Check that the url and config provided is valid, "
            + "and that the language in the url matches the config.\n"
        )


def flow_main(args):
    """Main dispatch for "flow" mode commands.

    :param args: arguments from parse_args()
    :type args: argparse.Namespace
    """
    conversation_config = get_conversation_config(args)
    settings = get_connection_settings(args)
    api = WebsocketClient(settings)
    transcripts = Transcripts()
    add_printing_handlers(
        api,
        transcripts,
        print_json=args["print_json"],
    )

    def run(stream):
        try:
            api.run_synchronously(
                [Interaction(stream)],
                get_audio_settings(args),
                conversation_config,
                from_cli=True,
            )
        except KeyboardInterrupt:
            # Gracefully handle Ctrl-C, else we get a huge stack-trace.
            LOGGER.warning("Keyboard interrupt received.")

    run(sys.stdin.buffer)


if __name__ == "__main__":
    main()



"""Define a set of typed dictionaries to represent a structured message
for a client side function, using TypedDict for type enforcement.

Example:
    tool_function = ToolFunctionParam(
        type="function",
        function={
            "name": "add_nums",
            "description": "Adds two numbers",
            "parameters": {
                "type": "object",
                "properties": {
                    "a": {"type": "int", "description": "First number"},
                    "b": {"type": "int", "description": "Second number"}
                },
                "required": ["a", "b"]
            }
        }
    )

    # Convert to dictionary
    dict(tool_function)
"""

import sys

from typing import Literal, Optional, Dict, List, TypedDict

if sys.version_info < (3, 11):
    from typing_extensions import Required
else:
    from typing import Required


class Property(TypedDict):
    type: Required[str]
    description: Required[str]


class FunctionParam(TypedDict, total=False):
    type: Required[str]
    properties: Required[Dict[str, Property]]
    required: Optional[List[str]]


class FunctionDefinition(TypedDict, total=False):
    name: Required[str]
    """The name of the function to be called."""

    description: Optional[str]
    """The description of what the function does, used by the model to choose
    when and how to call the function.
    """

    parameters: Optional[FunctionParam]
    """The parameters of the function to be called."""


class ToolFunctionParam(TypedDict):
    type: Required[Literal["function"]]
    """Currently, only 'function' is supported."""

    function: Required[FunctionDefinition]



"""Pre-configured Template IDs to start a Flow conversation."""

from enum import Enum
from typing import Literal


class Template(Enum):
    default = "default"
    amelia = "flow-service-assistant-amelia"
    humphrey = "flow-service-assistant-humphrey"


TemplateID = Literal[
    Template.default.value,
    Template.amelia.value,
    Template.humphrey.value,
]

# Map user-friendly name to full TemplateID
TemplateOptions = {t.name: t.value for t in Template}



from pytest import mark, param

from speechmatics_flow.tool_function_param import ToolFunctionParam


@mark.parametrize(
    "params",
    [
        param(
            {
                "type": "function",
                "function": {
                    "name": "test_function",
                    "description": "test_description",
                },
            },
            id="function without optional params",
        ),
        param(
            {
                "type": "function",
                "function": {
                    "name": "test_function",
                    "description": "test_description",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "a": {
                                "type": "int",
                                "description": "First number",
                            },
                            "b": {
                                "type": "int",
                                "description": "Second number",
                            },
                        },
                        "required": ["a", "b"],
                    },
                },
            },
            id="function with optional params",
        ),
    ],
)
def test_websocket_function(params):
    websocket_function = ToolFunctionParam(**params)
    assert dict(websocket_function) == params



# (c) 2024, Cantab Research Ltd.
"""
Data models and message types used by the library.
"""

import io
import ssl
from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Callable, Dict, Optional

from speechmatics_flow.templates import TemplateID


@dataclass
class AudioSettings:
    """Defines audio parameters."""

    encoding: str = "pcm_s16le"
    """Encoding format when raw audio is used. Allowed values are
    `pcm_f32le` and `pcm_s16le`"""

    sample_rate: int = 16000
    """Sampling rate in hertz."""

    chunk_size: int = 256
    """Chunk size in bytes."""

    def asdict(self):
        return {
            "type": "raw",
            "encoding": self.encoding,
            "sample_rate": self.sample_rate,
        }


@dataclass
class ConnectionSettings:
    """Defines connection parameters."""

    url: str
    """Websocket server endpoint."""

    message_buffer_size: int = 512
    """Message buffer size in bytes."""

    ssl_context: ssl.SSLContext = field(default_factory=ssl.create_default_context)
    """SSL context."""

    semaphore_timeout_seconds: float = 120
    """Semaphore timeout in seconds."""

    ping_timeout_seconds: float = 60
    """Ping-pong timeout in seconds."""

    auth_token: Optional[str] = None
    """auth token to authenticate a customer."""

    generate_temp_token: Optional[bool] = True
    """Automatically generate a temporary token for authentication."""


@dataclass
class ConversationConfig:
    """Defines configuration parameters for conversation requests."""

    template_id: TemplateID = "default"
    """Name of a predefined template."""

    template_variables: Optional[Dict[str, str]] = None
    """Optional parameter to allow overriding the default values of variables defined in the template."""

    def asdict(self):
        return asdict(
            self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )


class ClientMessageType(str, Enum):
    # pylint: disable=invalid-name
    """Defines various messages sent from client to server."""

    StartConversation = "StartConversation"
    """Initiates a conversation job based on configuration set previously."""

    AddAudio = "AddAudio"
    """Implicit name for all outbound binary messages. The server confirms
    receipt by sending an :py:attr:`ServerMessageType.AudioAdded` message."""

    AudioReceived = "AudioReceived"
    """Client response to :py:attr:`ServerMessageType.AddAudio`, indicating
    that audio has been added successfully."""

    AudioEnded = "AudioEnded"
    """Indicates audio input has finished."""

    ToolResult = "ToolResult"
    """Client response to :py:attr:`ServerMessageType.ToolInvoke`, containing
    the result of the function call."""


class ServerMessageType(str, Enum):
    # pylint: disable=invalid-name
    """Defines various message types sent from server to client."""

    ConversationStarted = "ConversationStarted"
    """Server response to :py:attr:`ClientMessageType.StartConversation`,
    acknowledging that a conversation session has started."""

    AddPartialTranscript = "AddPartialTranscript"
    """Indicates a partial transcript, which is an incomplete transcript that
    is immediately produced and may change as more context becomes available.
    """

    AudioAdded = "AudioAdded"
    """Server response to :py:attr:`ClientMessageType.AddAudio`, indicating
    that audio has been added successfully."""

    AddTranscript = "AddTranscript"
    """Indicates the final transcript of a part of the audio."""

    ResponseStarted = "ResponseStarted"
    """
    Indicates the start of TTS audio streaming from the server.
    The message contains the textual content of the utterance to be spoken.
    """

    ResponseCompleted = "ResponseCompleted"
    """Indicates the completion of TTS audio transmission from the server.
    The message includes the textual content of the utterance just spoken.
    """

    ResponseInterrupted = "ResponseInterrupted"
    """Indicates an interruption in the TTS audio stream from the server.
    The message contains the textual content up to the point where the utterance was stopped.
    """

    AddAudio = "AddAudio"
    """Implicit name for all outbound binary messages. The client confirms
    receipt by sending an :py:attr:`ClientMessageType.AudioReceived` message."""

    audio = "audio"
    """Message contains binary data"""

    prompt = "prompt"
    """Message contains text data"""

    ConversationEnding = "ConversationEnding"
    """Indicates the session will continue in one-sided mode
    during TTS playback of the final words."""

    ConversationEnded = "ConversationEnded"
    """Message indicates the session ended."""

    ToolInvoke = "ToolInvoke"
    """Indicates invocation of a function call. The client responds by sending
    an :py:attr:`ClientMessageType.ToolResult` message.
    """

    Info = "Info"
    """Indicates a generic info message."""

    Warning = "Warning"
    """Indicates a generic warning message."""

    Error = "Error"
    """Indicates n generic error message."""


@dataclass
class Interaction:
    """
    Defines a single interaction between a client and a server, typically
    used to handle non-continuous streams such as an audio file. This class
    enables the server to respond after the stream has finished or based
    on the specified callback function, allowing flexibility in connection
    handling after streaming.

    Attributes:
        stream (io.BufferedReader): The audio stream to be sent to the server.
        callback (Optional[Callable]): An optional function to be executed when
            the audio stream ends. This can be used to delay connection closure
            or perform additional actions upon stream completion.

    Examples:
        Keep the connection open for an additional 2 seconds after streaming
        an audio file, allowing time for the server to respond.

        ```python
        Interaction(audio_stream, callback=lambda x: time.sleep(2))
        ```
    """

    stream: io.BufferedReader
    """The audio stream to be sent to the server."""

    callback: Optional[Callable] = None
    """An optional function to be executed when the audio stream ends."""



import json
from typing import List, Dict, Optional

import pytest
from pytest import param

from speechmatics_flow.client import WebsocketClient
from speechmatics_flow.models import (
    ServerMessageType,
    ConnectionSettings,
    ClientMessageType,
    AudioSettings,
    ConversationConfig,
)
from speechmatics_flow.tool_function_param import ToolFunctionParam

TOOL_FUNCTION = dict(
    ToolFunctionParam(
        type="function",
        function={
            "name": "test_function",
            "description": "test function to be called.",
        },
    )
)


@pytest.fixture
def ws_client():
    return WebsocketClient(
        connection_settings=ConnectionSettings(url="ws://test"),
    )


@pytest.mark.parametrize(
    "audio_format, conversation_config, tools, expected_start_message",
    [
        param(
            AudioSettings(),
            ConversationConfig(),
            None,
            {
                "message": ClientMessageType.StartConversation.value,
                "audio_format": AudioSettings().asdict(),
                "conversation_config": ConversationConfig().asdict(),
            },
            id="with default values",
        ),
        param(
            AudioSettings(),
            ConversationConfig(),
            [TOOL_FUNCTION],
            {
                "message": ClientMessageType.StartConversation.value,
                "audio_format": AudioSettings().asdict(),
                "conversation_config": ConversationConfig().asdict(),
                "tools": [TOOL_FUNCTION],
            },
            id="with default values and tools",
        ),
    ],
)
def test_start_conversation(
    ws_client: WebsocketClient,
    audio_format: AudioSettings,
    conversation_config: ConversationConfig,
    tools: Optional[List[ToolFunctionParam]],
    expected_start_message: Dict,
):
    handler_called = False

    def handler(*_):
        nonlocal handler_called
        handler_called = True

    ws_client.middlewares = {ClientMessageType.StartConversation: [handler]}
    ws_client.audio_settings = audio_format
    ws_client.conversation_config = conversation_config
    ws_client.tools = tools
    start_conversation_msg = ws_client._start_conversation()
    assert start_conversation_msg == json.dumps(
        expected_start_message
    ), f"expected={start_conversation_msg}, got={expected_start_message}"
    assert handler_called, "handler was not called"


@pytest.mark.asyncio
async def test_consumer_supports_sync_and_async_handlers(ws_client):
    await ws_client._init_synchronization_primitives()

    async_handler_called = False
    sync_handler_called = False

    async def async_handler(_):
        nonlocal async_handler_called
        async_handler_called = True

    def sync_handler(_):
        nonlocal sync_handler_called
        sync_handler_called = True

    # Add event handlers for a message type
    ws_client.event_handlers = {
        ServerMessageType.ConversationStarted: [async_handler, sync_handler],
    }

    message = json.dumps({"message": ServerMessageType.ConversationStarted})
    await ws_client._consumer(message, from_cli=False)

    # Check if both handlers were called
    assert async_handler_called, "async handler was not called"
    assert sync_handler_called, "sync handler was not called"



from pytest import mark, param

from speechmatics_flow import cli
from speechmatics_flow.templates import Template

TEMPLATE_VARS = {
    "persona": "You are an English butler named Humphrey.",
    "style": "Be charming but unpredictable.",
    "context": "You are taking a customer's order at a fast food restaurant.",
}


@mark.parametrize(
    "args, exp_values",
    [
        param(
            [],
            {"template_id": Template.default.value},
            id="default assistant",
        ),
        param(
            ["--assistant=amelia"],
            {"template_id": Template.amelia.value},
            id="assistant amelia",
        ),
        param(
            ["--assistant=humphrey"],
            {"template_id": Template.humphrey.value},
            id="assistant humphrey",
        ),
        param(
            ["--config-file=tests/data/conversation_config.json"],
            {
                "template_id": "flow-service-assistant-humphrey",
                "template_variables": TEMPLATE_VARS,
            },
            id="params from config file",
        ),
        param(
            ["--assistant=amelia", "--config-file=tests/data/conversation_config.json"],
            {
                "template_id": "flow-service-assistant-amelia",
                "template_variables": TEMPLATE_VARS,
            },
            id="params from config file with assistant override from cli",
        ),
    ],
)
def test_get_conversation_config(args, exp_values):
    test_values = vars(cli.parse_args(args=args))
    config = cli.get_conversation_config(test_values)
    assert config.asdict() == exp_values, "Expecting {} but got {}".format(
        exp_values, config.asdict()
    )
